from importlib.metadata import entry_points

from karabo.bound import (
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    STRING_ELEMENT,
    VECTOR_BOOL_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    Schema,
    Timestamp,
    Trainstamp,
)

from TrainMatcher import TrainMatcher

from ._version import version as deviceVersion
from . import utils

my_schema = Schema()
(
    NODE_ELEMENT(my_schema)
    .key("data")
    .commit(),

    VECTOR_BOOL_ELEMENT(my_schema)
    .key("data.dataFramePattern")
    .assignmentOptional()
    .defaultValue([])
    .commit(),
)


kernels = [
    kernel_class.load()
    for kernel_class in entry_points().get("calng.arbiter_kernel", [])
]
kernel_choice = {
    kernel_class.__name__: kernel_class for kernel_class in kernels
}


@KARABO_CLASSINFO("FrameSelectionArbiter", deviceVersion)
class FrameSelectionArbiter(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            NODE_ELEMENT(expected)
            .key("frameSelection")
            .displayedName("Frame selection")
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameSelection.kernelChoice")
            .displayedName("Kernel to use")
            .assignmentMandatory()
            .options(",".join(kernel_choice.keys()))
            .reconfigurable()
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("output")
            .dataSchema(my_schema)
            .commit(),

            NODE_ELEMENT(expected)
            .key("frameSelection.kernels")
            .displayedName("Kernels")
            .commit(),
        )
        for kernel_class in kernel_choice.values():
            kernel_class.extend_device_schema(
                expected, f"frameSelection.kernels.{kernel_class._node_name}"
            )

    def initialization(self):
        super().initialization()
        self._kernel_class = kernel_choice[self.get("frameSelection.kernelChoice")]
        self._kernel = self._kernel_class(
            self._parameters[f"frameSelection.kernels.{self._kernel_class._node_name}"],
            self,
        )

    def on_matched_data(self, train_id, sources):
        # TODO: robust frame deduction
        for (data, _) in sources.values():
            if not data.has("image.cellId"):
                continue
            num_frames = data.get("image.cellId").size
            break

        decision = self._kernel.consider(train_id, sources, num_frames)
        result = Hash()
        # TODO: avoid recasting
        result["data.dataFramePattern"] = list(map(bool, decision))
        self.output.write(
            result,
            ChannelMetaData(
                f"{self.getInstanceId()}:output",
                Timestamp(Epochstamp(), Trainstamp(train_id)),
            ),
        )
        self.output.update()
        self.rate_out.update()

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        self._reinstantiate_arbiter_kernel = False
        if conf.has("frameSelection.kernelChoice"):
            self.log.INFO("Switching frame selection kernel")
            self._kernel_class = kernel_choice[conf.get("frameSelection.kernelChoice")]
            self._reinstantiate_arbiter_kernel = True
        elif conf.has(f"frameSelection.kernels.{self._kernel_class._node_name}"):
            self.log.INFO("Reconfiguring frame selection kernel")
            # TODO: update instead of rebuild for complex kernels
            # (decide based on reconfigurability?)
            self._reinstantiate_arbiter_kernel = True

    def postReconfigure(self):
        super().postReconfigure()
        if self._reinstantiate_arbiter_kernel:
            self._kernel = self._kernel_class(
                self._parameters[
                    f"frameSelection.kernels.{self._kernel_class._node_name}"
                ]
            )
            self._reinstantiate_arbiter_kernel = False
