import concurrent.futures
import contextlib
import enum
import functools
import gc
from importlib.metadata import entry_points
import math
import pathlib
import threading
from timeit import default_timer

import dateutil.parser
import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    INPUT_CHANNEL,
    INT32_ELEMENT,
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    VECTOR_STRING_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    ChannelMetaData,
    Hash,
    MetricPrefix,
    PythonDevice,
    Schema,
    State,
    Timestamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES

from . import scenes, preview_utils, shmem_utils, schemas, utils
from ._version import version as deviceVersion

PROCESSING_STATE_TIMEOUT = 10


class FramefilterSpecType(enum.Enum):
    NONE = "none"
    RANGE = "range"
    COMMASEPARATED = "commaseparated"


class WarningLampType(enum.Enum):
    FRAME_FILTER = enum.auto()
    MEMORY_CELL_RANGE = enum.auto()
    CONSTANT_OPERATING_PARAMETERS = enum.auto()
    PREVIEW_SETTINGS = enum.auto()
    CORRECTION_RUNNER = enum.auto()
    OUTPUT_BUFFER = enum.auto()
    GPU_MEMORY = enum.auto()
    CALCAT_CONNECTION = enum.auto()
    EMPTY_HASH = enum.auto()
    MISC_INPUT_DATA = enum.auto()
    TRAIN_ID = enum.auto()
    TIMESERVER_CONNECTION = enum.auto()


@KARABO_CLASSINFO("BaseCorrection", deviceVersion)
class BaseCorrection(PythonDevice):
    _available_addons = []  # classes, filled by add_addon_nodes using entry_points
    _base_output_schema = None  # subclass must set constructor
    _constant_enum_class = None  # subclass must set
    _correction_flag_class = None  # subclass must set (ex.: dssc_gpu.CorrectionFlags)
    _correction_steps = None  # subclass must set
    _kernel_runner_class = None  # subclass must set (ex.: dssc_gpu.DsscGpuRunner)
    _kernel_runner_init_args = {}  # optional extra args for runner
    _image_data_path = "image.data"  # customize for *some* subclasses
    _cell_table_path = "image.cellId"
    _pulse_table_path = "image.pulseId"
    _warn_memory_cell_range = True  # can be disabled for some detectors
    _cuda_pin_buffers = False

    def _load_constant_to_runner(self, constant, constant_data):
        """Subclass must define how to process constants into correction maps and store
        into appropriate buffers in (GPU or main) memory."""
        # note: aim to refactor this away as kernel runner handles loading logic
        raise NotImplementedError()

    def _successfully_loaded_constant_to_runner(self, constant):
        for field_name in self._constant_to_correction_names[constant]:
            key = f"corrections.{field_name}.available"
            if not self.unsafe_get(key):
                self.set(key, True)

        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to runner")

    @property
    def input_data_shape(self):
        """Subclass must define expected input data shape in terms of dataFormat.{
        memoryCells,pixelsX,pixelsY} and any other axes."""
        raise NotImplementedError()

    @property
    def output_data_shape(self):
        """Shape of corrected image data sent on dataOutput. Depends on data format
        parameters (optionally including frame filter)."""
        axis_lengths = {
            "x": self.unsafe_get("dataFormat.pixelsX"),
            "y": self.unsafe_get("dataFormat.pixelsY"),
            "f": self.unsafe_get("dataFormat.filteredFrames"),
        }
        return tuple(
            axis_lengths[axis] for axis in self.unsafe_get("dataFormat.outputAxisOrder")
        )

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
    ):
        """Subclass must define data processing (presumably using the kernel runner).
        Will be called by input_handler, which will take care of some common checks and
        extracting the parameters given to process_data."""
        raise NotImplementedError()

    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("state")
            .setNewDefaultValue(State.INIT)
            .commit(),

            INPUT_CHANNEL(expected).key("dataInput").commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("fastSources")
            .displayedName("Fast data sources")
            .description(
                "Sources to get data from. Only incoming hashes from these sources "
                "will be processed. This will typically be a single entry of the form: "
                "'[instrument]_DET_[detector]/DET/[channel]:xtdf'."
            )
            .assignmentOptional()
            .defaultValue([])
            .commit(),

            NODE_ELEMENT(expected)
            .key("frameFilter")
            .displayedName("Frame filter")
            .description(
                "The frame filter - if set - slices the input data. Frames not in the "
                "filter will be discarded before any processing happens and will not "
                "get to dataOutput or preview. Note that this filter goes by frame "
                "index rather than cell ID or pulse ID; set accordingly. Handle with "
                "care - an invalid filter can prevent all processing. How the filter "
                "is specified depends on frameFilter.type. See frameFilter.current to "
                "inspect the currently set frame filter array (if any)."
            )
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameFilter.type")
            .tags("managed")
            .displayedName("Type")
            .description(
                "Controls how frameFilter.spec is used. The default value of 'none' "
                "means that no filter is set (regardless of frameFilter.spec). "
                "'arange' allows between one and three integers separated by ',' which "
                "are parsed and passed directly to numpy.arange. 'commaseparated' "
                "reads a list of integers separated by commas."
            )
            .options(",".join(spectype.value for spectype in FramefilterSpecType))
            .assignmentOptional()
            .defaultValue("none")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameFilter.spec")
            .tags("managed")
            .displayedName("Specification")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("frameFilter.current")
            .displayedName("Current filter")
            .description(
                "This read-only value is used to display the contents of the current "
                "frame filter. An empty array means no filtering is done."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            UINT32_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .tags("managed")
            .displayedName("Output buffer size limit")
            .unit(Unit.BYTE)
            .metricPrefix(MetricPrefix.GIGA)
            .description(
                "Corrected trains are written to shared memory locations. These are "
                "pre-allocated and re-used (circular buffer). This parameter "
                "determines how much memory to set aside for that buffer."
            )
            .assignmentOptional()
            .defaultValue(10)
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("availableScenes")
            .setSpecialDisplayType(DT_SCENES)
            .readOnly()
            .initialValue(["overview", "constant_overrides"])
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useShmemHandles")
            .tags("managed")
            .displayedName("Use shared memory")
            .description(
                "If enabled, shared memory handles will be used to avoid copying "
                "the main image data on the dataOutput channel.  Note that these "
                "handles need to be dereferenced (typically by a group matcher) to "
                "get the data off the correction node."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useInfiniband")
            .tags("managed")
            .displayedName("Use infiniband")
            .description(
                "If enabled, device will during initialization try to bind its main "
                "data output channel (dataOutput) to its node's infiniband interface. "
                "Default interface is used if no 'ib0' interface is found."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("dataFormat")
            .displayedName("Data format (in/out)")
            .commit(),

            STRING_ELEMENT(expected)
            .key("dataFormat.inputImageDtype")
            .displayedName("Input image data dtype")
            .description("The (numpy) dtype to expect for incoming image data.")
            .readOnly()
            .initialValue("uint16")
            .commit(),

            STRING_ELEMENT(expected)
            .key("dataFormat.outputImageDtype")
            .tags("managed")
            .displayedName("Output image data dtype")
            .description(
                "The (numpy) dtype to use for outgoing image data. Input is cast to "
                "float32, corrections are applied, and only then will the result be "
                "cast to outputImageDtype. Be aware that casting to integer type "
                "causes truncation rather than rounding."
            )
            # TODO: consider adding rounding / binning for integer output
            .options("float16,float32,uint16")
            .assignmentOptional()
            .defaultValue("float32")
            .reconfigurable()
            .commit(),

            # important: determines shape of data as going into correction
            UINT32_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .displayedName("Pixels x")
            .description("Number of pixels of image data along X axis")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),

            UINT32_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .displayedName("Pixels y")
            .description("Number of pixels of image data along Y axis")
            .assignmentOptional()
            .defaultValue(128)
            .commit(),

            UINT32_ELEMENT(expected)
            .key("dataFormat.frames")
            .displayedName("Frames")
            .description("Number of image frames per train in incoming data")
            .assignmentOptional()
            .defaultValue(1)  # subclass will want to set a default value
            .commit(),

            UINT32_ELEMENT(expected)
            .key("dataFormat.filteredFrames")
            .displayedName("Frames after filter")
            .description("Number of frames left after applying frame filter")
            .readOnly()
            .initialValue(0)
            .commit(),

            STRING_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .tags("managed")
            .displayedName("Output axis order")
            .description(
                "Axes of main data output can be reordered after correction. Axis "
                "order is specified as string consisting of 'x', 'y', and 'f', with "
                "the latter indicating the image frame. The default value of 'fxy' "
                "puts pixels on the fast axes."
            )
            .options("fxy,fyx,xfy,xyf,yfx,yxf")
            .assignmentOptional()
            .defaultValue("fxy")
            .reconfigurable()
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.inputDataShape")
            .displayedName("Input data shape")
            .description(
                "Image data shape in incoming data (from reader / DAQ). This value is "
                "computed from pixelsX, pixelsY, and frames - this field just "
                "shows what is currently expected."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.outputDataShape")
            .displayedName("Output data shape")
            .description(
                "Image data shape for data output from this device. This value is "
                "computed from pixelsX, pixelsY, and the size of the frame filter - "
                "this field just shows what is currently expected."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.cellId")
            .readOnly()
            .initialValue([])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.pulseId")
            .readOnly()
            .initialValue([])
            .commit(),

            NODE_ELEMENT(expected)
            .key("workarounds")
            .displayedName("Workarounds")
            .commit(),

            BOOL_ELEMENT(expected)
            .key("workarounds.overrideInputAxisOrder")
            .tags("managed")
            .displayedName("Override input axis order")
            .description(
                "The shape of the image data ndarray as received from the "
                "DataAggregator is sometimes wrong - the axes are actually in a "
                "different order than the ndarray shape suggests. If this flag is on, "
                "the shape of the ndarray will be overridden with the axis order which "
                "was expected."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(expected)
            .key("workarounds.trainFromFutureThreshold")
            .tags("managed")
            .displayedName("Spurious future train ID threshold")
            .description(
                "Some detectors occasionally send a train with incorrect and much too "
                "large train ID. To avoid these 'future trains' from interfering with "
                "train matching, use this threshold to discard them immediately. If a "
                "train arrives with an ID which exceeds the current train ID from the "
                "time server by more than this threshold, the train is ignored."
            )
            .assignmentOptional()
            .defaultValue(10000)
            .reconfigurable()
            .commit(),
        )

        (
            SLOT_ELEMENT(expected)
            .key("loadMostRecentConstants")
            .tags("managed")
            .displayedName("Load most recent constants")
            .description(
                "Calling this slot will flush all constant buffers and cause the "
                "device to start querying CalCat for the most recent constants (all "
                "constants applicable for this device) available with the currently "
                "set constant parameters. This should typically be called after "
                "instantiating pipeline, after changing parameters, or after "
                "generating new constants."
            )
            .commit()
        )

        (
            STRING_ELEMENT(expected)
            .key("inputDataState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("NORMAL")
            .commit(),

            STRING_ELEMENT(expected)
            .key("deviceInternalsState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("NORMAL")
            .commit(),

            STRING_ELEMENT(expected)
            .key("processingState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("NORMAL")
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("preview")
            .displayedName("Preview")
            .commit(),

            INT32_ELEMENT(expected)
            .key("preview.index")
            .tags("managed")
            .displayedName("Index (or stat) for preview")
            .description(
                "If this value is ≥ 0, the corresponding index (frame, cell, or pulse) "
                "will be sliced for the preview output. If this value is < 0, preview "
                "will be one of the following stats: -1: max, -2: mean, -3: sum, -4: "
                "stdev. These stats are computed across frames, ignoring NaN values."
            )
            .assignmentOptional()
            .defaultValue(0)
            .minInc(-4)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("preview.selectionMode")
            .tags("managed")
            .displayedName("Index selection mode")
            .description(
                "The value of preview.index can be used in multiple ways, controlled "
                "by this value. If this is set to 'frame', preview.index is sliced "
                "directly from data. If 'cell' (or 'pulse') is selected, I will look "
                "at cell (or pulse) table for the requested cell (or pulse ID). "
                "Special (stat) index values <0 are not affected by this."
            )
            .options(
                ",".join(
                    spectype.value for spectype in utils.PreviewIndexSelectionMode
                )
            )
            .assignmentOptional()
            .defaultValue("frame")
            .reconfigurable()
            .commit(),
        )

        # just measurements and counters to display
        (
            UINT64_ELEMENT(expected)
            .key("trainId")
            .displayedName("Train ID")
            .description("ID of latest train processed by this device.")
            .readOnly()
            .initialValue(0)
            .commit(),

            NODE_ELEMENT(expected)
            .key("performance")
            .displayedName("Performance measures")
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(100)
            .info("Processing too slow to reach 10 Hz")
            .needsAcknowledging(False)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.rate")
            .displayedName("Rate")
            .description(
                "Actual rate with which this device gets, processes, and sends trains. "
                "This is a simple windowed moving average."
            )
            .unit(Unit.HERTZ)
            .readOnly()
            .initialValue(0)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.inputDelay")
            .displayedName("Input delay")
            .description(
                "Measured delay on the input channel. Measured by difference between "
                "'now' and timestamp in input metadata."
            )
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.ratioOfRecentTrainsReceived")
            .description(
                "Of the latest trains (from last received train, going back "
                "[some buffer range]), how many did we receive? This estimate is "
                "updated when new trains come in, so is unreliable if nothing is "
                "coming at all."
            )
            .unit(Unit.PERCENT)
            .readOnly()
            .initialValue(0)
            .commit(),
        )

        # this node will be filled out by subclass
        (
            NODE_ELEMENT(expected)
            .key("corrections")
            .displayedName("Correction steps")
            .commit(),
        )
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("warningLamps")
            .assignmentOptional()
            .defaultValue(
                [
                    "deviceInternalsState",
                    "inputDataState",
                    "processingState",
                ]
            )
            .commit()
        )
        # subclasses can populate with parameters / controls for plugins
        (
            NODE_ELEMENT(expected)
            .key("addons")
            .displayedName("Addons")
            .commit(),
        )

    def __init__(self, config):
        if config.get("useInfiniband", default=True):
            from PipeToZeroMQ.utils import find_infiniband_ip

            ib_ip = find_infiniband_ip()
            config["dataOutput.hostname"] = ib_ip
        super().__init__(config)

        self.output_data_dtype = np.dtype(config["dataFormat.outputImageDtype"])

        self.sources = set(config.get("fastSources"))

        self.kernel_runner = None  # must call _update_buffers to initialize
        self._shmem_buffer = None  # ditto
        self._use_shmem_handles = config.get("useShmemHandles")
        self._preview_friend = None

        self._correction_flag_enabled = self._correction_flag_class.NONE
        self._correction_flag_preview = self._correction_flag_class.NONE
        self._correction_applied_hash = Hash()
        # note: does not handle one constant enabling multiple correction steps
        # (do we need to generalize "decide if this correction step is available"?)
        self._constant_to_correction_names = {}
        for (name, _, constants) in self._correction_steps:
            for constant in constants:
                self._constant_to_correction_names.setdefault(constant, set()).add(name)
        self._buffer_lock = threading.Lock()
        self._last_processing_started = 0  # used for processing time and timeout
        self._last_train_id_processed = 0  # used to keep track (and as fallback)
        self._last_cell_table = None  # used to check whether to update property
        self._last_pulse_table = None  # ditto - though not all detectors have this

        # register slots
        def constant_override_fun(friend_fun, constant, preserve_fields):
            def aux():
                self.flush_constants(
                    constants={constant}, preserve_fields=preserve_fields
                )
                with self.warning_context(
                    f"foundConstants.{constant.name}.state", on_success="ON"
                ) as warn:
                    try:
                        constant_data = getattr(self.calcat_friend, friend_fun)(
                            constant
                        )
                        self._load_constant_to_runner(constant, constant_data)
                    except Exception as e:
                        warn(f"Failed to load {constant}: {e}")
                    else:
                        self._successfully_loaded_constant_to_runner(constant)
                # note: consider if multi-constant buffers need special care
            self._lock_and_update(aux)

        for constant in self._constant_enum_class:
            self.KARABO_SLOT(
                functools.partial(
                    constant_override_fun,
                    friend_fun="get_constant_version",
                    constant=constant,
                    preserve_fields=set(),
                ),
                slotName=f"foundConstants_{constant.name}_loadMostRecent",
                numArgs=0,
            )
            self.KARABO_SLOT(
                functools.partial(
                    constant_override_fun,
                    friend_fun="get_constant_from_constant_version_id",
                    constant=constant,
                    preserve_fields={"constantVersionId"},
                ),
                slotName=f"foundConstants_{constant.name}_overrideConstantFromVersion",
                numArgs=0,
            )
            self.KARABO_SLOT(
                functools.partial(
                    constant_override_fun,
                    friend_fun="get_constant_from_file",
                    constant=constant,
                    preserve_fields={"dataFilePath", "dataSetName"},
                ),
                slotName=f"foundConstants_{constant.name}_overrideConstantFromFile",
                numArgs=0,
            )

        self.KARABO_SLOT(self.loadMostRecentConstants)
        self.KARABO_SLOT(self.requestScene)

        self.registerInitialFunction(self._initialization)

    @contextlib.contextmanager
    def warning_context(
        self,
        schema_key,
        warn_type=None,
        on_success="NORMAL",
        on_error="ERROR",
        reraise=True,
        only_print_once=False,
    ):
        tracker = self._warning_trackers[schema_key]
        warn_fun = tracker.new_context(warn_type, only_print_once)
        try:
            yield warn_fun
        except Exception as e:
            warn_fun(f"Exception happened for {schema_key}, {warn_type}: {e}")
            if reraise:
                raise e
        finally:
            tracker.update_state(on_success=on_success, on_error=on_error)

    def _initialization(self):
        self.updateState(State.INIT)
        self._warning_trackers = {
            key: utils.ContextWarningLamp(self, key)
            for key in self.getFullSchema().getDefaultValue("warningLamps")
        }
        self._preview_friend = preview_utils.PreviewFriend(
            self,
            output_channels=self._preview_outputs,
        )
        self["availableScenes"] = self["availableScenes"] + [
            f"preview:{channel}" for channel in self._preview_outputs
        ]
        for constant in self._constant_enum_class:
            key = f"foundConstants.{constant.name}.state"
            self._warning_trackers[key] = utils.ContextWarningLamp(self, key)

        with self.warning_context(
            "deviceInternalsState", WarningLampType.CALCAT_CONNECTION
        ) as warn:
            try:
                self.calcat_friend = self._calcat_friend_class(
                    self, pathlib.Path.cwd() / "calibration-client-secrets.json"
                )
            except Exception as e:
                warn(f"Failed to connect to CalCat: {e}")
                # TODO: add raw fallback mode if CalCat fails (raw data still useful)
                return

        with self.warning_context(
            "processingState", WarningLampType.FRAME_FILTER
        ) as warn:
            try:
                self._frame_filter = _parse_frame_filter(self._parameters)
            except (ValueError, TypeError):
                warn("Failed to parse initial frame filter, will not use")
                self._frame_filter = None
        # TODO: decide how persistent initial warning should be
        self._update_correction_flags()
        self._update_frame_filter()

        self._buffered_status_update = Hash()
        self._processing_time_tracker = utils.ExponentialMovingAverage(alpha=0.3)
        self._rate_tracker = utils.WindowRateTracker()
        self._input_delay_tracker = utils.ExponentialMovingAverage(alpha=0.3)
        self._rate_update_timer = utils.RepeatingTimer(
            interval=1,
            callback=self._update_rate_and_state,
        )
        self._train_ratio_tracker = utils.TrainRatioTracker()

        self.KARABO_ON_INPUT("dataInput", self.input_handler)
        self.KARABO_ON_EOS("dataInput", self.handle_eos)

        self._enabled_addons = [
            addon_class(self._parameters[f"addons.{addon_class._node_name}"])
            for addon_class in self._available_addons
            if self.get(f"addons.{addon_class._node_name}.enable")
        ]
        if self._enabled_addons:
            schema_override = Schema()
            output_schema_override = self._base_output_schema
            for addon in self._enabled_addons:
                addon.extend_output_schema(output_schema_override)
            (
                # add channel to schema override to re-inject with new output schema
                OUTPUT_CHANNEL(schema_override)
                .key("dataOutput")
                .dataSchema(output_schema_override)
                .commit(),
            )
            # update device schema in one fell swoop
            self.updateSchema(schema_override)

        self.updateState(State.ON)

    def __del__(self):
        del self._shmem_buffer
        super().__del__()

    def preReconfigure(self, config):
        # validation
        for ts_path in (
            "constantParameters.deviceMappingSnapshotAt",
            "constantParameters.constantVersionEventAt",
        ):
            if config.has(ts_path):
                ts_string = config.get(ts_path)
                if ts_string.strip() == "":
                    config.set(ts_path, "")
                else:
                    # allow exception to reach operator
                    timestamp = dateutil.parser.isoparse(ts_string)
                    config.set(ts_path, timestamp.isoformat())

        if config.has("constantParameters.deviceMappingSnapshotAt"):
            self.calcat_friend.flush_pdu_mapping()

        # update device based on changes
        if config.has("frameFilter"):
            self._frame_filter = _parse_frame_filter(
                utils.ChainHash(config, self._parameters)
            )

        self._prereconfigure_update_hash = config

    def postReconfigure(self):
        if not hasattr(self, "_prereconfigure_update_hash"):
            self.log_status_warn("postReconfigure without knowing update hash")
            return

        update = self._prereconfigure_update_hash

        if update.has("frameFilter"):
            self._lock_and_update(self._update_frame_filter)
        elif any(
            update.has(shape_param)
            for shape_param in (
                "dataFormat.pixelsX",
                "dataFormat.pixelsY",
                "dataFormat.outputImageDtype",
                "dataFormat.outputAxisOrder",
                "dataFormat.frames",
                "constantParameters.memoryCells",
                "frameFilter",
            )
        ):
            self._lock_and_update(self._update_buffers)

        if any(
            (
                update.has(f"corrections.{field_name}.enable")
                or update.has(f"corrections.{field_name}.preview")
            )
            for field_name, *_ in self._correction_steps
        ):
            self._update_correction_flags()

        # TODO: only send subhash of reconfiguration (like with addons)
        if update.has("preview"):
            self._preview_friend.reconfigure(update)

        if update.has("addons"):
            # note: can avoid iterating, but it's not that costly
            for addon in self._enabled_addons:
                full_path = f"addons.{addon._node_name}"
                if update.has(full_path):
                    addon.reconfigure(update[full_path])

    def _lock_and_update(self, method, background=True):
        # TODO: securely handle errors (postReconfigure may succeed, device state not)
        def runner():
            with self._buffer_lock, utils.StateContext(self, State.CHANGING):
                method()

        if background:
            threading.Thread(target=runner, daemon=True).start()
        else:
            runner()

    def loadMostRecentConstants(self):
        def aux():
            self.flush_constants()
            # TODO: ignore irrelevant constants (like agpid thresholding in fixed gain)
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future_to_constant = {
                    executor.submit(
                        self.calcat_friend.get_constant_version,
                        constant,
                    ): constant
                    for constant in self._constant_enum_class
                }
                for future in concurrent.futures.as_completed(future_to_constant):
                    constant = future_to_constant[future]
                    with self.warning_context(
                        f"foundConstants.{constant.name}.state", on_success="ON"
                    ) as warn:
                        try:
                            constant_data = future.result()
                            self._load_constant_to_runner(constant, constant_data)
                        except Exception as e:
                            warn(f"Failed to load {constant}: {e}")
                        else:
                            self._successfully_loaded_constant_to_runner(constant)
        self._lock_and_update(aux)

    def flush_constants(self, constants=None, preserve_fields=None):
        """Reset constant buffers; the correction maps (often buffers on GPU) should be
        set to identity values and constant caches cleared. By default do this for all
        constants - a subset can instead be specified via the constants parameter
        (should be iterable of constant enum values).

        After flushing, the kernel runner will be instructed to reload any remaining
        cached constants after flushing. This can take a second.

        The parameter preserve_fields is passed to the CalCat friend; this is used for
        overrides; in those cases, we want to get value from an overridden field after
        flushing everything else, so don't wipe it out.
        """
        if constants is None:
            constants = set(self._constant_enum_class)
        else:
            constants = set(constants)
        self.kernel_runner.flush_buffers(constants)
        for field_name, _, used_constants in self._correction_steps:
            if (
                used_constants
                and None not in used_constants
                and constants >= used_constants
            ):
                self.set(f"corrections.{field_name}.available", False)
        self.calcat_friend.flush_constants(constants, preserve_fields)
        self._reload_constants_from_cache_to_runner(constants)
        self._update_correction_flags()

    def log_status_info(self, msg):
        self.log.INFO(msg)
        self.set("status", msg)

    def log_status_warn(self, msg):
        self.log.WARN(msg)
        self.set("status", msg)

    def requestScene(self, params):
        payload = Hash()
        name = params.get("name", default="")
        payload["name"] = name
        payload["success"] = True
        if name == "overview":
            payload["data"] = scenes.correction_device_overview(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
            )
        elif name == "constant_overrides":
            payload["data"] = scenes.correction_device_constant_overrides(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
            )
        elif name.startswith("preview:"):
            channel_name = name[len("preview:") :]
            payload["data"] = scenes.correction_device_preview(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
                preview_channel=channel_name,
            )
        elif name.startswith("browse_schema"):
            if ":" in name:
                prefix = name[len("browse_schema:") :]
            else:
                prefix = "managed"
            payload["data"] = scenes.recursive_subschema_scene(
                self.getInstanceId(),
                self.getFullSchema(),
                prefix,
            )
        else:
            payload["success"] = False
        response = Hash()
        response["type"] = "deviceScene"
        response["origin"] = self.getInstanceId()
        response["payload"] = payload
        self.reply(response)

    def _write_output(self, data, old_metadata):
        """For dataOutput: reusing incoming data hash and setting source and timestamp
        to be same as input"""
        metadata = ChannelMetaData(
            old_metadata.get("source"),
            Timestamp.fromHashAttributes(old_metadata.getAttributes("timestamp")),
        )
        data["corrections"] = self._correction_applied_hash

        channel = self.signalSlotable.getOutputChannel("dataOutput")
        channel.write(data, metadata, copyAllData=False)
        channel.update()

    def _update_correction_flags(self):
        """Based on constants loaded and settings, update bit mask flags for kernel"""
        enabled = self._correction_flag_class.NONE
        output = Hash()  # for informing downstream receivers what was applied
        preview = self._correction_flag_class.NONE
        for field_name, flag, constants in self._correction_steps:
            output[field_name] = False
            if self.get(f"corrections.{field_name}.available"):
                if self.get(f"corrections.{field_name}.enable"):
                    enabled |= flag
                    output[field_name] = True
                if self.get(f"corrections.{field_name}.preview"):
                    preview |= flag
        self._correction_flag_enabled = enabled
        self._correction_flag_preview = preview
        self._correction_applied_hash = output
        self.log.DEBUG(f"Corrections for dataOutput: {str(enabled)}")
        self.log.DEBUG(f"Corrections for preview: {str(preview)}")

    def _update_frame_filter(self, update_buffers=True):
        """Parse frameFilter string (if set) and update cached filter array. May update
        dataFormat.filteredFrames - will therefore by default call _update_buffers
        afterwards."""
        # TODO: add some validation to preReconfigure
        self.log.DEBUG("Updating frame filter")

        if self._frame_filter is None:
            self.set("dataFormat.filteredFrames", self.get("dataFormat.frames"))
            self.set("frameFilter.current", [])
        else:
            self.set("dataFormat.filteredFrames", self._frame_filter.size)
            self.set("frameFilter.current", list(map(int, self._frame_filter)))

        with self.warning_context(
            "deviceInternalsState", WarningLampType.FRAME_FILTER
        ) as warn:
            if self._frame_filter is not None and (
                self._frame_filter.min() < 0
                or self._frame_filter.max() >= self.get("dataFormat.frames")
            ):
                warn("Invalid frame filter set, expect exceptions!")
                # note: dataFormat.frames could change from detector and
                # that could make this warning invalid

        if update_buffers:
            self._update_buffers()

    def _update_buffers(self):
        """(Re)initialize buffers / kernel runner according to expected data shapes"""
        self.log.INFO("Updating buffers according to data shapes")
        # reflect the axis reordering in the expected output shape
        self.set("dataFormat.inputDataShape", list(self.input_data_shape))
        self.set("dataFormat.outputDataShape", list(self.output_data_shape))
        self.log.INFO(f"Input shape: {self.input_data_shape}")
        self.log.INFO(f"Output shape: {self.output_data_shape}")

        if self._shmem_buffer is None:
            shmem_buffer_name = self.getInstanceId() + ":dataOutput"
            memory_budget = self.get("outputShmemBufferSize") * 2**30
            self.log.INFO(f"Opening new shmem buffer: {shmem_buffer_name}")
            self._shmem_buffer = shmem_utils.ShmemCircularBuffer(
                memory_budget,
                self.output_data_shape,
                self.output_data_dtype,
                shmem_buffer_name,
            )
            self._shmem_receiver = shmem_utils.ShmemCircularBufferReceiver()
            if self._cuda_pin_buffers:
                self.log.INFO("Trying to pin the shmem buffer memory")
                self._shmem_buffer.cuda_pin()
            self.log.INFO("Done, shmem buffer is ready")
        else:
            self._shmem_buffer.change_shape(self.output_data_shape)

        # give CuPy a chance to at least start memory cleanup before this
        if self.kernel_runner is not None:
            del self.kernel_runner
            self.kernel_runner = None
            gc.collect()

        self.kernel_runner = self._kernel_runner_class(
            self.get("dataFormat.pixelsX"),
            self.get("dataFormat.pixelsY"),
            self.get("dataFormat.filteredFrames"),
            int(self.get("constantParameters.memoryCells")),
            output_data_dtype=self.output_data_dtype,
            **self._kernel_runner_init_args,
        )

        self._reload_constants_from_cache_to_runner()

    def _reload_constants_from_cache_to_runner(self, constants=None):
        # TODO: gracefully handle change in constantParameters.memoryCells
        if constants is None:
            constants = set(self._constant_enum_class)
        else:
            constants = set(constants)

        with self.calcat_friend.cached_constants_lock:
            for (
                constant,
                data,
            ) in self.calcat_friend.cached_constants.items():
                if constant not in constants:
                    continue
                with self.warning_context(
                    f"foundConstants.{constant.name}.state", on_success="ON"
                ) as warn:
                    try:
                        self.log_status_info(f"Reload constant {constant.name}")
                        self._load_constant_to_runner(constant, data)
                    except Exception as e:
                        warn(f"Failed to reload {constant.name}: {e}")
                    else:
                        self._successfully_loaded_constant_to_runner(constant)

    def input_handler(self, input_channel):
        """Main handler for data input: Do a few simple checks to determine whether to
        even try processing. If yes, will pass data and information to process_data
        method provided by subclass."""

        # Is device even ready for this?
        state = State[self.unsafe_get("state")]
        if state is State.ERROR:
            # in this case, we should have already issued warning
            return

        with self.warning_context(
            "deviceInternalsState", WarningLampType.CORRECTION_RUNNER
        ) as warn:
            if self.kernel_runner is None or state is State.INIT:
                warn("Received data before correction device was ready")
                return

        all_metadata = input_channel.getMetaData()
        for input_index in range(input_channel.size()):
            with self.warning_context(
                "inputDataState", WarningLampType.MISC_INPUT_DATA
            ) as warn:
                data_hash = input_channel.read(input_index)
                metadata = all_metadata[input_index]
                source = metadata.get("source")

                if source not in self.sources:
                    continue
                elif not data_hash.has(self._image_data_path):
                    warn("Ignoring hash without image node")
                    continue

            with self.warning_context(
                "inputDataState",
                WarningLampType.EMPTY_HASH,
                only_print_once=True,
            ) as warn:
                self._shmem_receiver.dereference_shmem_handles(data_hash)
                try:
                    image_data = np.asarray(data_hash.get(self._image_data_path))
                    cell_table = (
                        np.asarray(data_hash.get(self._cell_table_path)).ravel()
                        if self._cell_table_path is not None
                        else None
                    )
                    pulse_table = (
                        np.asarray(data_hash.get(self._pulse_table_path)).ravel()
                        if self._pulse_table_path is not None
                        else None
                    )
                except RuntimeError as err:
                    warn(
                        "Failed to load image data; "
                        f"probably empty hash from DAQ: {err}"
                    )
                    self._maybe_update_cell_and_pulse_tables(None, None)
                    continue

            # no more common reasons to skip input, so go to processing
            self._last_processing_started = default_timer()
            if state is State.ON:
                self.updateState(State.PROCESSING)
                self.log_status_info("Processing data")

            self._maybe_update_cell_and_pulse_tables(cell_table, pulse_table)
            timestamp = Timestamp.fromHashAttributes(metadata.getAttributes("timestamp"))
            train_id = timestamp.getTrainId()

            # check time server connection
            with self.warning_context(
                "deviceInternalsState", WarningLampType.TIMESERVER_CONNECTION
            ) as warn:
                my_timestamp = self.getActualTimestamp()
                my_train_id = my_timestamp.getTrainId()
                self._input_delay_tracker.update(
                    (my_timestamp.toTimestamp() - timestamp.toTimestamp()) * 1000
                )
                self._buffered_status_update.set(
                    "performance.inputDelay", self._input_delay_tracker.get()
                )
                if my_train_id == 0:
                    my_train_id = self._last_train_id_processed + 1
                    warn(
                        "Failed to get current train ID, using previously seen train "
                        "ID for future train thresholding - if this persists, check "
                        "connection to timeserver."
                    )
            with self.warning_context(
                "inputDataState", WarningLampType.TRAIN_ID, only_print_once=True
            ) as warn:
                if train_id > (
                    my_train_id
                    + self.unsafe_get("workarounds.trainFromFutureThreshold")
                ):
                    warn(
                        f"Suspecting train from the future: 'now' is {my_train_id}, "
                        f"received train ID {train_id}, dropping data"
                    )
                    continue
                try:
                    self._train_ratio_tracker.update(train_id)
                    self._buffered_status_update.set(
                        "performance.ratioOfRecentTrainsReceived",
                        self._train_ratio_tracker.get(
                            current_train=my_train_id,
                            expected_delay=math.ceil(
                                self.unsafe_get("performance.inputDelay") / 100
                            ),
                        )
                        if my_train_id != 0
                        else self._train_ratio_tracker.get(),
                    )
                except utils.NonMonotonicTrainIdWarning as ex:
                    warn(
                        f"Train ratio tracker noticed issue with train ID: {ex}\n"
                        f"For the record, I think now is: {my_train_id}"
                    )
                    self._train_ratio_tracker.reset()
                    self._train_ratio_tracker.update(train_id)

            if self._warn_memory_cell_range:
                with self.warning_context(
                    "processingState", WarningLampType.MEMORY_CELL_RANGE
                ) as warn:
                    if (
                        self.unsafe_get("constantParameters.memoryCells")
                        <= cell_table.max()
                    ):
                        warn("Input cell IDs out of range of constants")

            # TODO: avoid branching multiple times on cell_table
            if cell_table is not None and cell_table.size != self.unsafe_get(
                "dataFormat.frames"
            ):
                self.log_status_info(
                    f"Updating new input shape to account for {cell_table.size} cells"
                )
                self.set("dataFormat.frames", cell_table.size)
                self._lock_and_update(self._update_frame_filter, background=False)

            # DataAggregator typically tells us the wrong axis order
            if self.unsafe_get("workarounds.overrideInputAxisOrder"):
                expected_shape = self.input_data_shape
                if expected_shape != image_data.shape:
                    image_data.shape = expected_shape

            with self._buffer_lock:
                self.process_data(
                    data_hash,
                    metadata,
                    source,
                    train_id,
                    image_data,
                    cell_table,
                    pulse_table,
                )
            self._last_train_id_processed = train_id
            self._buffered_status_update.set("trainId", train_id)
            self._processing_time_tracker.update(
                default_timer() - self._last_processing_started
            )
            self._buffered_status_update.set(
                "performance.processingTime", self._processing_time_tracker.get() * 1000
            )
            self._rate_tracker.update()

    def _update_rate_and_state(self):
        if self.get("state") is State.PROCESSING:
            # always update rate: estimate depends on query time
            self._buffered_status_update.set(
                "performance.rate", self._rate_tracker.get()
            )
            # remaining stats are set _buffered_status_update in input handler
            self.set(self._buffered_status_update)
            self._buffered_status_update.clear()
            if (
                default_timer() - self._last_processing_started
                > PROCESSING_STATE_TIMEOUT
            ):
                self.updateState(State.ON)
                self.log_status_info(
                    f"No new train in {PROCESSING_STATE_TIMEOUT} s, switching state."
                )
                self._train_ratio_tracker.reset()
                self._maybe_update_cell_and_pulse_tables(None, None)

    def _maybe_update_cell_and_pulse_tables(self, cell_table, pulse_table):
        if not np.array_equal(cell_table, self._last_cell_table):
            self._last_cell_table = cell_table
            if cell_table is None:
                self.set("dataFormat.cellId", [])
            else:
                self.set("dataFormat.cellId", list(map(int, cell_table)))
        if not np.array_equal(pulse_table, self._last_pulse_table):
            self._last_pulse_table = pulse_table
            if pulse_table is None:
                self.set("dataFormat.pulseId", [])
            else:
                self.set("dataFormat.pulseId", list(map(int, pulse_table)))

    def handle_eos(self, channel):
        self.updateState(State.ON)
        self.signalEndOfStream("dataOutput")


utils.add_unsafe_get(BaseCorrection)


def add_correction_step_schema(schema, field_flag_constants_mapping):
    """Using the fields in the provided mapping, will add nodes to schema

    field_flag_mapping is assumed to be iterable of pairs where first entry in each
    pair is the name of a correction step as it will appear in device schema (second
    entry - typically an enum field - is ignored). For correction step, a node and some
    booleans are added to the schema (all tagged as managed). Subclass can customize /
    add additional keys under node later.

    This method should be called in expectedParameters of subclass after the same for
    BaseCorrection has been called. Would be nice to include in BaseCorrection instead,
    but that is tricky: static method of superclass will need _correction_steps
    of subclass or device server gets mad. A nice solution with classmethods would be
    welcome.
    """

    for field_name, _, used_constants in field_flag_constants_mapping:
        node_name = f"corrections.{field_name}"
        (
            NODE_ELEMENT(schema).key(node_name).commit(),

            BOOL_ELEMENT(schema)
            .key(f"{node_name}.available")
            .displayedName("Available")
            .description(
                "This boolean indicates whether the necessary constants have been "
                "loaded for this correction step to be applied. Enabling the "
                "correction will have no effect unless this is True."
            )
            .readOnly()
            # some corrections available without constants
            .initialValue(not used_constants or None in used_constants)
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{node_name}.enable")
            .tags("managed")
            .displayedName("Enable")
            .description(
                "Controls whether to apply this correction step for main data "
                "output - subject to availability."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{node_name}.preview")
            .tags("managed")
            .displayedName("Preview")
            .description(
                "Whether to apply this correction step for corrected preview "
                "output - subject to availability."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
        )


def add_bad_pixel_config_node(schema, prefix="corrections.badPixels"):
    (
        STRING_ELEMENT(schema)
        .key("corrections.badPixels.maskingValue")
        .tags("managed")
        .displayedName("Bad pixel masking value")
        .description(
            "Any pixels masked by the bad pixel mask will have their value replaced "
            "with this. Note that this parameter is to be interpreted as a "
            "numpy.float32; use 'nan' to get NaN value."
        )
        .assignmentOptional()
        .defaultValue("nan")
        .reconfigurable()
        .commit(),

        NODE_ELEMENT(schema)
        .key("corrections.badPixels.subsetToUse")
        .displayedName("Bad pixel flags to use")
        .description(
            "The booleans under this node allow for selecting a subset of bad pixel "
            "types to take into account when doing bad pixel masking. Upon updating "
            "these flags, the map used for bad pixel masking will be ANDed with this "
            "selection. Turning disabled flags back on causes reloading of cached "
            "constants."
        )
        .commit(),
    )
    for field in utils.BadPixelValues:
        (
            BOOL_ELEMENT(schema)
            .key(f"corrections.badPixels.subsetToUse.{field.name}")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit()
        )


def add_preview_outputs(schema, channels):
    for channel in channels:
        (
            OUTPUT_CHANNEL(schema)
            .key(f"preview.{channel}")
            .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=True))
            .commit(),
        )
    preview_utils.PreviewFriend.add_schema(schema, output_channels=channels)


def add_addon_nodes(schema, device_class, prefix="addons"):
    det_name = device_class.__name__[:-len("Correction")].lower()
    device_class._available_addons = [
        addon.load()
        for addon in entry_points().get("calng.correction_addon", [])
        if not addon.extras
        or det_name in (extra[0] for extra in addon.extras)
    ]

    for addon_class in device_class._available_addons:
        (
            NODE_ELEMENT(schema)
            .key(f"{prefix}.{addon_class._node_name}")
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{prefix}.{addon_class._node_name}.enable")
            .assignmentOptional()
            .defaultValue(False)
            .commit(),
        )
        addon_class.extend_device_schema(
            schema, f"{prefix}.{addon_class._node_name}"
        )


def get_bad_pixel_field_selection(self):
    selection = 0
    for field in utils.BadPixelValues:
        if self.get(f"corrections.badPixels.subsetToUse.{field.name}"):
            selection |= field
    return selection


def _parse_frame_filter(config):
    filter_type = FramefilterSpecType(config["frameFilter.type"])
    filter_string = config["frameFilter.spec"]

    if filter_type is FramefilterSpecType.NONE or filter_string.strip() == "":
        return None
    elif filter_type is FramefilterSpecType.RANGE:
        # allow exceptions
        numbers = tuple(int(part) for part in filter_string.split(","))
        return np.arange(*numbers, dtype=np.uint16)
    elif filter_type is FramefilterSpecType.COMMASEPARATED:
        # np.fromstring is too lenient I think
        return np.array([int(s) for s in filter_string.split(",")], dtype=np.uint16)
    else:
        raise TypeError(f"Unknown frame filter type {filter_type}")
