import functools
import pathlib

from karabo.bound import (
    FLOAT_ELEMENT,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    UINT32_ELEMENT,
)
from .base_addon import BaseCorrectionAddon
from .. import utils


class Peakfinder9Addon(BaseCorrectionAddon):
    _node_name = "peakfinder9"

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            UINT32_ELEMENT(schema)
            .key(f"{prefix}.windowRadius")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(2)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{prefix}.maxPeaks")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(500)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.minPeakValueOverNeighbors")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(10)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.minSnrMaxPixel")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.minSnrPeakPixels")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(4)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.minSnrWholePeak")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(6)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{prefix}.minSigma")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{prefix}.blockX")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{prefix}.blockY")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{prefix}.blockZ")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(64)
            .reconfigurable()
            .commit(),
        )

    @staticmethod
    def extend_output_schema(schema):
        (
            NODE_ELEMENT(schema)
            .key("peakfinding")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("peakfinding.numPeaks")
            .dtype("UINT32")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("peakfinding.peakX")
            .dtype("FLOAT")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("peakfinding.peakY")
            .dtype("FLOAT")
            .commit(),

            NDARRAY_ELEMENT(schema)
            .key("peakfinding.peakIntensity")
            .dtype("FLOAT")
            .commit(),
        )

    def post_correction(self, processed_data, cell_table, pulse_table, output_hash):
        # assumes processed data shape is frames, pixels, pixels
        if self._input_shape != processed_data.shape:
            try:
                del self._peakfinding_parameters
            except AttributeError:
                pass
            try:
                del self._grid_and_block
            except AttributeError:
                pass
            self._input_shape = processed_data.shape
            self._rebuild_buffers()
        kernel_params = self._peakfinding_parameters  # this will create buffers
        self._peak_counts.fill(0)
        self.kernel(
            *self._grid_and_block,
            (
                *kernel_params,
                processed_data.astype(cupy.float32, copy=False),
                self._peak_counts,
                self._peak_x,
                self._peak_y,
                self._peak_intensity,
            ),
        )
        output_hash.set("peakfinding.numPeaks", self._peak_counts.get())
        output_hash.set("peakfinding.peakX", self._peak_x.get())
        output_hash.set("peakfinding.peakY", self._peak_y.get())
        output_hash.set("peakfinding.peakIntensity", self._peak_intensity.get())

    def reconfigure(self, changed_config):
        self._config.merge(changed_config)
        try:
            del self._peakfinding_parameters
        except AttributeError:
            pass
        if changed_config.has("maxPeaks"):
            self._rebuild_buffers
        if any(changed_config.has(key) for key in ("blockX", "blockY", "blockZ")):
            try:
                del self._grid_and_block
            except AttributeError:
                pass

    def __init__(self, config):
        global cupy
        import cupy

        self._config = config

        _src_dir = pathlib.Path(__file__).absolute().parent.parent
        with (_src_dir / "kernels" / "peakfinder9_gpu.cu").open("r") as fd:
            self.kernel = cupy.RawKernel(
                code=fd.read(),
                name="pf9",
                options=("--std=c++11",),
                backend="nvcc",
            )

        self._input_shape = (0, 0, 0)  # frames, ss, fs

    @functools.cached_property
    def _grid_and_block(self):
        # TODO: optimize
        arbitrary_block_shape = (
            self._config["blockX"],
            self._config["blockY"],
            self._config["blockZ"],
        )
        return (
            utils.grid_to_cover_shape_with_blocks(
                self._input_shape, arbitrary_block_shape
            ),
            arbitrary_block_shape,
        )

    @functools.cached_property
    def _peakfinding_parameters(self):
        return (
            cupy.uint16(self._input_shape[0]),
            cupy.uint16(self._input_shape[1]),
            cupy.uint16(self._input_shape[2]),
            cupy.uint16(self._config["windowRadius"]),
            cupy.float32(self._config["minPeakValueOverNeighbors"]),
            cupy.float32(self._config["minSnrMaxPixel"]),
            cupy.float32(self._config["minSnrPeakPixels"]),
            cupy.float32(self._config["minSnrWholePeak"]),
            cupy.float32(self._config["minSigma"]),
            cupy.uint32(self._config["maxPeaks"]),
        )

    def _rebuild_buffers(self):
        output_shape = (self._input_shape[0], self._config["maxPeaks"])
        self._peak_counts = cupy.zeros(self._input_shape[0], dtype=cupy.uint32)
        self._peak_x = cupy.empty(output_shape, dtype=cupy.float32)
        self._peak_y = cupy.empty(output_shape, dtype=cupy.float32)
        self._peak_intensity = cupy.empty(output_shape, dtype=cupy.float32)
