import enum

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    VECTOR_STRING_ELEMENT,
    ChannelMetaData,
    Hash,
    Timestamp,
)

from .. import base_calcat, base_correction, base_kernel_runner, schemas, utils
from .._version import version as deviceVersion


_pretend_pulse_table = np.arange(2720, dtype=np.uint8)


class Constants(enum.Enum):
    LUTGotthard2 = enum.auto()
    OffsetGotthard2 = enum.auto()
    RelativeGainGotthard2 = enum.auto()
    BadPixelsDarkGotthard2 = enum.auto()
    BadPixelsFFGotthard2 = enum.auto()


bp_constant_types = {
    Constants.BadPixelsDarkGotthard2,
    Constants.BadPixelsFFGotthard2,
}


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    LUT = 1
    OFFSET = 2
    GAIN = 4
    BPMASK = 8


class Gotthard2CpuRunner(base_kernel_runner.BaseKernelRunner):
    _gpu_based = False
    _xp = np

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
        bad_pixel_subset=None,
    ):
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            output_data_dtype,
        )

        from ..kernels import gotthard2_cython

        self.correction_kernel = gotthard2_cython.correct
        self.input_shape = (frames, pixels_x)
        self.processed_shape = self.input_shape
        # model: 2 buffers (corresponding to actual memory cells), 2720 frames
        # lut maps from uint12 to uint10 values
        self.lut_shape = (2, 4096, pixels_x)
        self.map_shape = (3, self.constant_memory_cells, self.pixels_x)
        self.lut = np.empty(self.lut_shape, dtype=np.uint16)
        self.offset_map = np.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_map = np.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = np.empty(self.map_shape, dtype=np.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value
        self.flush_buffers()
        self._bad_pixel_subset = bad_pixel_subset

        self.input_data = None  # will just point to data coming in
        self.input_gain_stage = None  # will just point to data coming in
        self.processed_data = None  # will just point to buffer we're given

    @property
    def preview_data_views(self):
        return (self.input_data, self.processed_data)

    def load_constant(self, constant_type, data):
        if constant_type is Constants.LUTGotthard2:
            self.lut[:] = np.transpose(data.astype(np.uint16, copy=False), (1, 2, 0))
        elif constant_type is Constants.OffsetGotthard2:
            self.offset_map[:] = np.transpose(data.astype(np.float32, copy=False))
        elif constant_type is Constants.RelativeGainGotthard2:
            self.rel_gain_map[:] = np.transpose(data.astype(np.float32, copy=False))
        elif constant_type in bp_constant_types:
            # TODO: add the regular bad pixel subset configuration
            data = data.astype(np.uint32, copy=False)
            if self._bad_pixel_subset is not None:
                data = data & self._bad_pixel_subset
            self.bad_pixel_map |= np.transpose(data)
        else:
            raise ValueError(f"What is this constant '{constant_type}'?")

    def set_bad_pixel_subset(self, subset, apply_now=True):
        self._bad_pixel_subset = subset
        if apply_now:
            self.bad_pixel_map &= self._bad_pixel_subset

    def load_data(self, image_data, input_gain_stage):
        """Experiment: loading both in one function as they are tied"""
        self.input_data = image_data.astype(np.uint16, copy=False)
        self.input_gain_stage = input_gain_stage.astype(np.uint8, copy=False)

    def flush_buffers(self, constants=None):
        # for now, always flushing all here...
        default_lut = (
            np.arange(2 ** 12).astype(np.float64) * 2 ** 10 / 2 ** 12
        ).astype(np.uint16)
        self.lut[:] = np.stack([np.stack([default_lut] * 2)] * self.pixels_x, axis=2)
        self.offset_map.fill(0)
        self.rel_gain_map.fill(1)
        self.bad_pixel_map.fill(0)

    def correct(self, flags, out=None):
        if out is None:
            out = np.empty(self.processed_shape, dtype=np.float32)

        self.correction_kernel(
            self.input_data,
            self.input_gain_stage,
            np.uint8(flags),
            self.lut,
            self.offset_map,
            self.rel_gain_map,
            self.bad_pixel_map,
            self.bad_pixel_mask_value,
            out,
        )
        self.processed_data = out
        return out


class Gotthard2CalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {c: self.condition for c in Constants}

    def condition(self):
        res = base_calcat.OperatingConditions()
        res["Exposure time"] = self._get_param("exposureTime")
        res["Exposure period"] = self._get_param("exposurePeriod")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Single photon"] = self._get_param("singlePhoton")
        res["Acquisition rate"] = self._get_param("acquisitionRate")
        return res

    @staticmethod
    def add_schema(schema):
        super(Gotthard2CalcatFriend, Gotthard2CalcatFriend).add_schema(
            schema, "Gotthard2-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(1280)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(2)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(200)
            .commit()
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.exposureTime")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.exposurePeriod")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.singlePhoton")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.acquisitionRate")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(
            schema, Constants
        )


@KARABO_CLASSINFO("Gotthard2Correction", deviceVersion)
class Gotthard2Correction(base_correction.BaseCorrection):
    _base_output_schema = schemas.jf_output_schema(use_shmem_handle=False)
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        (
            "lut",
            CorrectionFlags.LUT,
            {Constants.LUTGotthard2},
        ),
        (
            "offset",
            CorrectionFlags.OFFSET,
            {Constants.OffsetGotthard2},
        ),
        (
            "gain",
            CorrectionFlags.GAIN,
            {Constants.RelativeGainGotthard2},
        ),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            {
                Constants.BadPixelsDarkGotthard2,
                Constants.BadPixelsFFGotthard2,
            }
        ),
    )
    _kernel_runner_class = Gotthard2CpuRunner
    _calcat_friend_class = Gotthard2CalcatFriend
    _constant_enum_class = Constants
    _image_data_path = "data.adc"
    _cell_table_path = "data.memoryCell"
    _pulse_table_path = None
    _warn_memory_cell_range = False  # for now, receiver always writes 255
    _preview_outputs = ["outputStreak"]
    _cuda_pin_buffers = False

    @staticmethod
    def expectedParameters(expected):
        super(Gotthard2Correction, Gotthard2Correction).expectedParameters(expected)
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(Gotthard2Correction._base_output_schema)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(1280)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(2720)  # note: actually just frames...
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),
        )

        base_correction.add_preview_outputs(
            expected, Gotthard2Correction._preview_outputs
        )
        for channel in ("outputRaw", "outputCorrected", "outputFrameSums"):
            # add these "manually" as the automated bits wrap ImageData
            (
                OUTPUT_CHANNEL(expected)
                .key(f"preview.{channel}")
                .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=False))
                .commit(),
            )
        base_correction.add_correction_step_schema(
            expected,
            Gotthard2Correction._correction_steps,
        )
        base_correction.add_bad_pixel_config_node(expected)
        Gotthard2CalcatFriend.add_schema(expected)

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            self.unsafe_get("dataFormat.pixelsX"),
        )

    @property
    def output_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            self.unsafe_get("dataFormat.pixelsX"),
        )

    @property
    def _kernel_runner_init_args(self):
        return {
            "bad_pixel_mask_value": self._bad_pixel_mask_value,
            "bad_pixel_subset": self._bad_pixel_subset,
        }

    @property
    def _bad_pixel_mask_value(self):
        return np.float32(self.unsafe_get("corrections.badPixels.maskingValue"))

    _bad_pixel_subset = property(base_correction.get_bad_pixel_field_selection)

    def postReconfigure(self):
        super().postReconfigure()
        update = self._prereconfigure_update_hash
        if update.has("corrections.badPixels.subsetToUse"):
            if any(
                update.get(
                    f"corrections.badPixels.subsetToUse.{field.name}", default=False
                )
                for field in utils.BadPixelValues
            ):
                self.log_status_info(
                    "Some fields reenabled, reloading cached bad pixel constants"
                )
                self.kernel_runner.set_bad_pixel_subset(
                    self._bad_pixel_subset, apply_now=False
                )
                with self.calcat_friend.cached_constants_lock:
                    self.kernel_runner.flush_buffers(bp_constant_types)
                    for constant_type in (
                        bp_constant_types & self.calcat_friend.cached_constants.keys()
                    ):
                        self._load_constant_to_runner(
                            constant_type,
                            self.calcat_friend.cached_constants[constant_type],
                        )
            else:
                # just narrowing the subset - no reload, just AND
                self.kernel_runner.set_bad_pixel_subset(
                    self._bad_pixel_subset, apply_now=True
                )
        if update.has("corrections.badPixels.maskingvalue"):
            self.kernel_runner.bad_pixel_mask_value = self._bad_pixel_mask_value

    def __init__(self, config):
        super().__init__(config)
        try:
            np.float32(config.get("corrections.badPixels.maskingValue"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        pulse_table,
    ):
        # cell table currently not used for GOTTHARD2 (assume alternating)
        gain_map = np.asarray(data_hash.get("data.gain"))
        if self.unsafe_get("workarounds.overrideInputAxisOrder"):
            gain_map.shape = self.input_data_shape
        try:
            self.kernel_runner.load_data(image_data, gain_map)
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.correct(self._correction_flag_enabled, out=buffer_array)

        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                _pretend_pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
        (
            preview_raw,
            preview_corrected,
        ) = self.kernel_runner.compute_previews(preview_slice_index)

        if self._use_shmem_handles:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        self._write_output(data_hash, metadata)

        frame_sums = np.nansum(buffer_array, axis=1)
        timestamp = Timestamp.fromHashAttributes(metadata.getAttributes("timestamp"))
        for channel, data in (
            ("outputRaw", preview_raw),
            ("outputCorrected", preview_corrected),
            ("outputFrameSums", frame_sums),
        ):
            output = self.signalSlotable.getOutputChannel(f"preview.{channel}")
            output.write(
                Hash(
                    "image.data",
                    data,
                    "image.mask",
                    (~np.isfinite(data)).astype(np.uint8),
                ),
                ChannelMetaData(f"{self.getInstanceId()}:preview.{channel}", timestamp),
            )
            output.update()
        self._preview_friend.write_outputs(metadata, buffer_array)

    def _load_constant_to_runner(self, constant, constant_data):
        self.kernel_runner.load_constant(constant, constant_data)
