import concurrent.futures
import enum
import functools

import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Schema,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    preview_utils,
    utils,
)
from .._version import version as deviceVersion


_pretend_pulse_table = np.arange(16, dtype=np.uint8)


class Constants(enum.Enum):
    Offset10Hz = enum.auto()
    BadPixelsDark10Hz = enum.auto()
    BadPixelsFF10Hz = enum.auto()
    RelativeGain10Hz = enum.auto()


bad_pixel_constants = {
    Constants.BadPixelsDark10Hz,
    Constants.BadPixelsFF10Hz,
}


# from pycalibration (TOOD: move to common shared lib)
class GainModes(enum.Enum):
    ADAPTIVE_GAIN = 0
    FIXED_GAIN = 1


class GainSettings(enum.Enum):
    LOW_CDS = 0
    HIGH_CDS = 1


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    REL_GAIN = 2
    BPMASK = 4
    STRIXEL = 8


class JungfrauBaseRunner(base_kernel_runner.BaseKernelRunner):
    _corrected_axis_order = "fyx"
    _xp = None  # subclass sets CuPy (import at runtime) or numpy

    @property
    def input_shape(self):
        return self.frames, self.pixels_y, self.pixels_x

    @property
    def processed_shape(self):
        return self.input_shape

    @property
    def map_shape(self):
        return (self.constant_memory_cells, self.pixels_y, self.pixels_x, 3)

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
    ):
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            output_data_dtype,
        )
        # note: superclass creates cell table with wrong dtype
        self.input_gain_stage = self._xp.empty(self.input_shape, dtype=np.uint8)
        self.offset_map = self._xp.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_map = self._xp.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = self._xp.empty(self.map_shape, dtype=np.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value

        # strixel support
        self._strixel_out_shape = (frames, 86, 3090)
        self._strixel_block = (1, 1, 64)
        # note: only executing kernel on lower half of y range, hence 256
        self._strixel_grid = tuple(
            utils.ceil_div(a_length, block_length)
            for (a_length, block_length) in zip(
                (frames, 256, 1024), self._strixel_block
            )
        )
        self._processed_data_regular = self._xp.empty(
            self.processed_shape, dtype=output_data_dtype
        )
        self._processed_data_strixel = self._xp.empty(
            self._strixel_out_shape, dtype=output_data_dtype
        )
        self.flush_buffers(set(Constants))

    def load_constant(self, constant, constant_data):
        if constant_data.shape[0] == self.pixels_x:
            constant_data = np.transpose(constant_data, (2, 1, 0, 3))
        else:
            constant_data = np.transpose(constant_data, (2, 0, 1, 3))

        if constant is Constants.Offset10Hz:
            self.offset_map[:] = self._xp.asarray(constant_data, dtype=np.float32)
        elif constant is Constants.RelativeGain10Hz:
            self.rel_gain_map[:] = self._xp.asarray(constant_data, dtype=np.float32)
        elif constant in bad_pixel_constants:
            self.bad_pixel_map |= self._xp.asarray(constant_data, dtype=np.uint32)
        else:
            raise ValueError(f"Unexpected constant type {constant}")

    @property
    def preview_data_views(self):
        return (self.input_data, self.processed_data, self.input_gain_stage)

    @property
    def burst_mode(self):
        return self.frames > 1

    def flush_buffers(self, constants):
        if Constants.Offset10Hz in constants:
            self.offset_map.fill(0)
        if Constants.RelativeGain10Hz in constants:
            self.rel_gain_map.fill(1)
        if constants & {
            Constants.BadPixelsDark10Hz,
            Constants.BadPixelsFF10Hz,
        }:
            self.bad_pixel_map.fill(0)
            self.bad_pixel_map[
                :, :, 255:1023:256
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map[
                :, :, 256:1024:256
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map[
                :, [255, 256]
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value

    def override_bad_pixel_flags_to_use(self, override_value):
        self.bad_pixel_map &= self._xp.uint32(override_value)


class JungfrauGpuRunner(JungfrauBaseRunner):
    _gpu_based = True

    def _pre_init(self):
        import cupy as cp

        self._xp = cp

    def _post_init(self):
        self.input_data = self._xp.empty(self.input_shape, dtype=np.uint16)
        self.cell_table = self._xp.empty(self.frames, dtype=np.uint8)
        self.block_shape = (1, 1, 64)
        self.grid_shape = utils.grid_to_cover_shape_with_blocks(
            self.input_shape, self.block_shape
        )

        source_module = self._xp.RawModule(
            code=base_kernel_runner.get_kernel_template("jungfrau_gpu.cu").render(
                {
                    "pixels_x": self.pixels_x,
                    "pixels_y": self.pixels_y,
                    "frames": self.frames,
                    "constant_memory_cells": self.constant_memory_cells,
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self.output_data_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                    "burst_mode": self.burst_mode,
                }
            )
        )
        self.correction_kernel = source_module.get_function("correct")
        self.strixel_transform_kernel = source_module.get_function("strixel_transform")

    def load_data(self, image_data, input_gain_stage, cell_table):
        self.input_data.set(image_data)
        self.input_gain_stage.set(input_gain_stage)
        if self.burst_mode:
            self.cell_table.set(cell_table)

    def correct(self, flags):
        self.correction_kernel(
            self.grid_shape,
            self.block_shape,
            (
                self.input_data,
                self.input_gain_stage,
                self.cell_table,
                self._xp.uint8(flags),
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self._processed_data_regular,
            ),
        )
        if flags & CorrectionFlags.STRIXEL:
            self.strixel_transform_kernel(
                self._strixel_grid,
                self._strixel_block,
                (
                    self._processed_data_regular,
                    self._processed_data_strixel,
                ),
            )
            self.processed_data = self._processed_data_strixel
        else:
            self.processed_data = self._processed_data_regular


class JungfrauCpuRunner(JungfrauBaseRunner):
    _gpu_based = False
    _xp = np

    def _post_init(self):
        # not actually allocating, will just point to incoming data
        self.input_data = None
        self.input_gain_stage = None
        self.input_cell_table = None

        # for computing previews faster
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=3)

        from ..kernels import jungfrau_cython

        self.correction_kernel_single = jungfrau_cython.correct_single
        self.correction_kernel_burst = jungfrau_cython.correct_burst
        self.correction_kernel_strixel = jungfrau_cython.strixel_transform

    def __del__(self):
        self.thread_pool.shutdown()

    def load_data(self, image_data, input_gain_stage, cell_table):
        self.input_data = image_data.astype(np.uint16, copy=False)
        self.input_gain_stage = input_gain_stage.astype(np.uint8, copy=False)
        self.input_cell_table = cell_table.astype(np.uint8, copy=False)

    def correct(self, flags):
        if self.burst_mode:
            self.correction_kernel_burst(
                self.input_data,
                self.input_gain_stage,
                self.input_cell_table,
                flags,
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self._processed_data_regular,
            )
        else:
            self.correction_kernel_single(
                self.input_data,
                self.input_gain_stage,
                flags,
                self.offset_map,
                self.rel_gain_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self._processed_data_regular,
            )

        if flags & CorrectionFlags.STRIXEL:
            self.correction_kernel_strixel(
                self._processed_data_regular, self._processed_data_strixel
            )
            self.processed_data = self._processed_data_strixel
        else:
            self.processed_data = self._processed_data_regular

    def compute_previews(self, preview_index):
        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.frames:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        if preview_index >= 0:

            def fun(a):
                return a[preview_index]

        elif preview_index == -1:
            # note: separate from next case because dtype not applicable here
            fun = functools.partial(np.nanmax, axis=0)
        elif preview_index in (-2, -3, -4):
            fun = functools.partial(
                {
                    -2: np.nanmean,
                    -3: np.nansum,
                    -4: np.nanstd,
                }[preview_index],
                axis=0,
                dtype=np.float32,
            )
        return self.thread_pool.map(fun, self.preview_data_views)


class JungfrauCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.Offset10Hz: self.dark_condition,
            Constants.BadPixelsDark10Hz: self.dark_condition,
            Constants.BadPixelsFF10Hz: self.dark_condition,
            Constants.RelativeGain10Hz: self.dark_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(JungfrauCalcatFriend, JungfrauCalcatFriend).add_schema(
            schema, "jungfrau-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(90)
            .commit(),
        )

        # add extra parameters
        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .tags("managed")
            .displayedName("Integration time")
            .description("Integration time in ms")
            .assignmentOptional()
            .defaultValue(350)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sensorTemperature")
            .tags("managed")
            .displayedName("Sensor temperature")
            .description("Sensor temperature in K")
            .assignmentOptional()
            .defaultValue(291)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainMode")
            .tags("managed")
            .displayedName("Gain mode")
            .description(
                "Detector may be operating in one of several gain modes. In CalCat, "
                "these map to two CalCat parameters: 'Gain Setting' which is 1 for "
                "dynamic gain with HG0 (0 otherwise) and 'Gain mode' which is 1 "
                "for fixed gain (omitted otherwise)."
            )
            .assignmentOptional()
            .defaultValue(GainModes.ADAPTIVE_GAIN.name)
            .options(",".join(gain_mode.name for gain_mode in GainModes))
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainSetting")
            .tags("managed")
            .displayedName("Gain setting")
            .description("See description of gainMode")
            .assignmentOptional()
            .defaultValue(GainSettings.LOW_CDS.name)
            .options(",".join(gain_setting.name for gain_setting in GainSettings))
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Integration Time"] = self._get_param("integrationTime")
        res["Sensor Temperature"] = self._get_param("sensorTemperature")

        if (
            gain_mode := GainModes[self._get_param("gainMode")]
        ) is not GainModes.ADAPTIVE_GAIN:
            # NOTE: currently only including if parameter for CalCat is 1
            # change if conditions are tidied up in the database
            res["Gain mode"] = gain_mode.value
        res["Gain Setting"] = GainSettings[self._get_param("gainSetting")].value

        return res


@KARABO_CLASSINFO("JungfrauCorrection", deviceVersion)
class JungfrauCorrection(base_correction.BaseCorrection):
    _base_output_schema = schemas.jf_output_schema()
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        ("offset", CorrectionFlags.OFFSET, {Constants.Offset10Hz}),
        ("relGain", CorrectionFlags.REL_GAIN, {Constants.RelativeGain10Hz}),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            {
                Constants.BadPixelsDark10Hz,
                Constants.BadPixelsFF10Hz,
                None,
            },
        ),
        ("strixel", CorrectionFlags.STRIXEL, set()),
    )
    _kernel_runner_class = None  # note: set in __init__ based on config
    _calcat_friend_class = JungfrauCalcatFriend
    _constant_enum_class = Constants
    _preview_outputs = [
        "outputRaw",
        "outputCorrected",
        "outputGainMap",
    ]
    _image_data_path = "data.adc"
    _cell_table_path = "data.memoryCell"
    _pulse_table_path = None

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(JungfrauCorrection._base_output_schema)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(1024)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),

            # JUNGFRAU data is small, can fit plenty of trains in here
            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),
        )
        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(base_kernel_runner.KernelRunnerTypes.GPU.name)
            .options(
                ",".join(
                    kernel_type.name
                    for kernel_type in base_kernel_runner.KernelRunnerTypes
                )
            )
            .reconfigurable()
            .commit(),
        )

        base_correction.add_preview_outputs(
            expected, JungfrauCorrection._preview_outputs
        )
        base_correction.add_correction_step_schema(
            expected,
            JungfrauCorrection._correction_steps,
        )
        (
            OVERWRITE_ELEMENT(expected)
            .key("corrections.strixel.enable")
            .setNewDefaultValue(False)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("corrections.strixel.preview")
            .setNewDefaultValue(False)
            .commit(),
        )
        base_correction.add_bad_pixel_config_node(expected)
        JungfrauCalcatFriend.add_schema(expected)

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            self.unsafe_get("dataFormat.pixelsY"),
            self.unsafe_get("dataFormat.pixelsX"),
        )

    @property
    def _warn_memory_cell_range(self):
        # disable warning in normal operation as cell 15 is expected
        return self.unsafe_get("dataFormat.frames") > 1

    @property
    def output_data_shape(self):
        if self._correction_flag_enabled & CorrectionFlags.STRIXEL:
            axis_lengths = {
                "x": 3090,
                "y": 86,
                "f": self.unsafe_get("dataFormat.filteredFrames"),
            }
            return tuple(
                axis_lengths[axis]
                for axis in self.unsafe_get("dataFormat.outputAxisOrder")
            )
        return super().output_data_shape

    @property
    def _kernel_runner_class(self):
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            self.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            return JungfrauCpuRunner
        else:
            return JungfrauGpuRunner

    @property
    def _kernel_runner_init_args(self):
        return {
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
        }

    @property
    def bad_pixel_mask_value(self):
        return np.float32(self.unsafe_get("corrections.badPixels.maskingValue"))

    _override_bad_pixel_flags = property(base_correction.get_bad_pixel_field_selection)

    def __init__(self, config):
        super().__init__(config)
        try:
            np.float32(config.get("corrections.badPixels.maskingValue", default="nan"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

        if config.get("useShmemHandles", default=True):
            schema_override = Schema()
            (
                OUTPUT_CHANNEL(schema_override)
                .key("dataOutput")
                .dataSchema(schemas.jf_output_schema(use_shmem_handle=False))
                .commit(),
            )
            self.updateSchema(schema_override)

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        pulse_table,
    ):
        if len(cell_table.shape) == 0:
            cell_table = cell_table[np.newaxis]
        try:
            gain_map = data_hash["data.gain"]
            if self.unsafe_get("workarounds.overrideInputAxisOrder"):
                gain_map.shape = self.input_data_shape
            self.kernel_runner.load_data(image_data, gain_map, cell_table)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )

        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                _pretend_pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
        (
            preview_raw,
            preview_corrected,
            preview_gain_map,
        ) = self.kernel_runner.compute_previews(preview_slice_index)

        # reusing input data hash for sending
        if self._use_shmem_handles:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            # TODO: use shmem for data.gain, too
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        self._write_output(data_hash, metadata)

        self._preview_friend.write_outputs(
            metadata, preview_raw, preview_corrected, preview_gain_map
        )

    def _load_constant_to_runner(self, constant, constant_data):
        if constant in bad_pixel_constants:
            constant_data &= self._override_bad_pixel_flags
        self.kernel_runner.load_constant(constant, constant_data)

    def postReconfigure(self):
        super().postReconfigure()

        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash

        if update.has("corrections.strixel.enable"):
            self._lock_and_update(self._update_frame_filter)

        if update.has("corrections.badPixels.subsetToUse"):
            self.log_status_info("Updating bad pixel maps based on subset specified")
            # note: now just always reloading from cache for convenience
            with self.calcat_friend.cached_constants_lock:
                self.kernel_runner.flush_buffers(bad_pixel_constants)
                for constant in bad_pixel_constants:
                    if constant in self.calcat_friend.cached_constants:
                        self._load_constant_to_runner(
                            constant, self.calcat_friend.cached_constants[constant]
                        )
