import enum

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Schema,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from .._version import version as deviceVersion


class Constants(enum.Enum):
    Offset = enum.auto()
    GainAmpMap = enum.auto()
    RelativeGain = enum.auto()
    FFMap = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsFF = enum.auto()


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    GAIN_AMP = 2
    REL_GAIN = 4
    FF_CORR = 8
    BPMASK = 16


class LpdGpuRunner(base_kernel_runner.BaseKernelRunner):
    _gpu_based = True
    _corrected_axis_order = "fyx"

    @property
    def input_shape(self):
        return (self.frames, 1, self.pixels_y, self.pixels_x)

    @property
    def processed_shape(self):
        return (self.frames, self.pixels_y, self.pixels_x)

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
    ):
        import cupy as cp
        self._xp = cp
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            output_data_dtype,
        )
        self.gain_map = cp.empty(self.processed_shape, dtype=np.float32)
        self.input_data = cp.empty(self.input_shape, dtype=np.uint16)
        self.processed_data = cp.empty(self.processed_shape, dtype=output_data_dtype)
        self.cell_table = cp.empty(frames, dtype=np.uint16)

        self.map_shape = (constant_memory_cells, pixels_y, pixels_x, 3)
        self.offset_map = cp.empty(self.map_shape, dtype=np.float32)
        self.gain_amp_map = cp.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_slopes_map = cp.empty(self.map_shape, dtype=np.float32)
        self.flatfield_map = cp.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = cp.empty(self.map_shape, dtype=np.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value
        self.flush_buffers(set(Constants))

        self.correction_kernel = cp.RawModule(
            code=base_kernel_runner.get_kernel_template("lpd_gpu.cu").render(
                {
                    "pixels_x": self.pixels_x,
                    "pixels_y": self.pixels_y,
                    "frames": self.frames,
                    "constant_memory_cells": self.constant_memory_cells,
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self.output_data_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                }
            )
        ).get_function("correct")

        self.block_shape = (1, 1, 64)
        self.grid_shape = utils.grid_to_cover_shape_with_blocks(
            self.processed_shape, self.block_shape
        )

    def load_data(self, image_data, cell_table):
        self.input_data.set(image_data)
        self.cell_table.set(cell_table)

    @property
    def preview_data_views(self):
        # TODO: always split off gain from raw to avoid messing up preview?
        return (
            self.input_data[:, 0],  # raw
            self.processed_data,  # corrected
            self.gain_map,  # gain (split from raw)
        )

    def correct(self, flags):
        self.correction_kernel(
            self.grid_shape,
            self.block_shape,
            (
                self.input_data,
                self.cell_table,
                self._xp.uint8(flags),
                self.offset_map,
                self.gain_amp_map,
                self.rel_gain_slopes_map,
                self.flatfield_map,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self.gain_map,
                self.processed_data,
            ),
        )

    def load_constant(self, constant_type, constant_data):
        # constant type → transpose order
        bad_pixel_loading = {
            Constants.BadPixelsDark: (2, 1, 0, 3),
            Constants.BadPixelsFF: (2, 0, 1, 3),
        }
        # constant type → transpose order, GPU buffer
        other_constant_loading = {
            Constants.Offset: ((2, 1, 0, 3), self.offset_map),
            Constants.GainAmpMap: ((2, 0, 1, 3), self.gain_amp_map),
            Constants.FFMap: ((2, 0, 1, 3), self.flatfield_map),
            Constants.RelativeGain: ((2, 0, 1, 3), self.rel_gain_slopes_map),
        }
        if constant_type in bad_pixel_loading:
            self.bad_pixel_map |= self._xp.asarray(
                constant_data.transpose(bad_pixel_loading[constant_type]),
                dtype=np.uint32,
            )[:self.constant_memory_cells]
        elif constant_type in other_constant_loading:
            transpose_order, gpu_buffer = other_constant_loading[constant_type]
            gpu_buffer.set(
                np.transpose(
                    constant_data.astype(np.float32),
                    transpose_order,
                )[:self.constant_memory_cells]
            )
        else:
            raise ValueError(f"Unhandled constant type {constant_type}")

    def flush_buffers(self, constants):
        if Constants.Offset in constants:
            self.offset_map.fill(0)
        if Constants.GainAmpMap in constants:
            self.gain_amp_map.fill(1)
        if Constants.RelativeGain in constants:
            self.rel_gain_slopes_map.fill(1)
        if Constants.FFMap in constants:
            self.flatfield_map.fill(1)
        if constants & {Constants.BadPixelsDark, Constants.BadPixelsFF}:
            self.bad_pixel_map.fill(0)


class LpdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.Offset: self.with_cell_condition,
            Constants.BadPixelsDark: self.with_cell_condition,
            Constants.GainAmpMap: self.category_condition,
            Constants.FFMap: self.category_condition,
            Constants.RelativeGain: self.category_condition,
            Constants.BadPixelsFF: self.category_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(LpdCalcatFriend, LpdCalcatFriend).add_schema(schema, "LPD-Type")

        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(250)
            .commit(),
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.feedbackCapacitor")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.photonEnergy")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(9.3)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.category")
            .tags("managed")
            .displayedName("Category")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key("constantParameters.useMemoryCellOrder")
            .tags("managed")
            .displayedName("Use memory cell order parameter")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.memoryCellOrder")
            .tags("managed")
            .displayedName("Memory cell order")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def basic_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Feedback capacitor"] = self._get_param("feedbackCapacitor")

        return res

    def with_cell_condition(self):
        res = self.basic_condition()
        if self._get_param("useMemoryCellOrder"):
            res["Memory cell order"] = self._get_param("memoryCellOrder")

        return res

    def illuminated_condition(self):
        res = self.basic_condition()
        res["Source Energy"] = self._get_param("photonEnergy")
        return res

    def category_condition(self):
        res = self.illuminated_condition()
        res["category"] = self._get_param("category")
        return res


@KARABO_CLASSINFO("LpdCorrection", deviceVersion)
class LpdCorrection(base_correction.BaseCorrection):
    _base_output_schema = schemas.xtdf_output_schema()
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        ("offset", CorrectionFlags.OFFSET, {Constants.Offset}),
        ("gainAmp", CorrectionFlags.GAIN_AMP, {Constants.GainAmpMap}),
        ("relGain", CorrectionFlags.REL_GAIN, {Constants.RelativeGain}),
        ("flatfield", CorrectionFlags.FF_CORR, {Constants.FFMap}),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            {
                Constants.BadPixelsDark,
                Constants.BadPixelsFF,
            }
        ),
    )
    _kernel_runner_class = LpdGpuRunner
    _calcat_friend_class = LpdCalcatFriend
    _constant_enum_class = Constants
    _preview_outputs = ["outputRaw", "outputCorrected", "outputGainMap"]

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(LpdCorrection._base_output_schema)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(512)
            .commit(),

            # TODO: determine ideal default
            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),
        )

        # TODO: add bad pixel config node
        LpdCalcatFriend.add_schema(expected)
        base_correction.add_addon_nodes(expected, LpdCorrection)
        base_correction.add_correction_step_schema(expected, LpdCorrection._correction_steps)
        base_correction.add_preview_outputs(expected, LpdCorrection._preview_outputs)

        # additional settings for correction steps
        (
            STRING_ELEMENT(expected)
            .key("corrections.badPixels.maskingValue")
            .tags("managed")
            .displayedName("Bad pixel masking value")
            .description(
                "Any pixels masked by the bad pixel mask will have their value "
                "replaced with this. Note that this parameter is to be interpreted as "
                "a numpy.float32; use 'nan' to get NaN value."
            )
            .assignmentOptional()
            .defaultValue("nan")
            .reconfigurable()
            .commit(),
        )

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            1,
            self.unsafe_get("dataFormat.pixelsY"),
            self.unsafe_get("dataFormat.pixelsX"),
        )

    def __init__(self, config):
        super().__init__(config)
        try:
            np.float32(config.get("corrections.badPixels.maskingValue"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

        if config.get("useShmemHandles", default=False):
            def aux():
                schema_override = Schema()
                (
                    OUTPUT_CHANNEL(schema_override)
                    .key("dataOutput")
                    .dataSchema(schemas.xtdf_output_schema(use_shmem_handle=False))
                    .commit(),
                )

            self.registerInitialFunction(aux)

    def _load_constant_to_runner(self, constant, constant_data):
        self.kernel_runner.load_constant(constant, constant_data)

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        pulse_table,
    ):
        with self.warning_context(
            "processingState", base_correction.WarningLampType.CONSTANT_OPERATING_PARAMETERS
        ) as warn:
            if (
                cell_table_string := utils.cell_table_to_string(cell_table)
            ) != self.unsafe_get(
                "constantParameters.memoryCellOrder"
            ) and self.unsafe_get(
                "constantParameters.useMemoryCellOrder"
            ):
                warn(f"Cell order does not match input; input: {cell_table_string}")
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            # drop the singleton dimension for kernel runner
            self.kernel_runner.load_data(image_data, cell_table)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.correct(self._correction_flag_enabled)
        for addon in self._enabled_addons:
            addon.post_correction(
                self.kernel_runner.processed_data, cell_table, pulse_table, data_hash
            )
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
            (
                preview_raw,
                preview_corrected,
                preview_gain_map,
            ) = self.kernel_runner.compute_previews(preview_slice_index)

        if self._use_shmem_handles:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
        data_hash.set("image.pulseId", pulse_table[:, np.newaxis])

        self._write_output(data_hash, metadata)
        self._preview_friend.write_outputs(
            metadata, preview_raw, preview_corrected, preview_gain_map
        )

    def preReconfigure(self, config):
        # TODO: DRY (taken from AGIPD device)
        super().preReconfigure(config)
        if config.has("corrections.badPixels.maskingValue"):
            # only check if it is valid (let raise exception)
            # if valid, postReconfigure will use it
            np.float32(config.get("corrections.badPixels.maskingValue"))

    def postReconfigure(self):
        super().postReconfigure()
        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash
        if update.has("corrections.badPixels.maskingValue"):
            self.kernel_runner.bad_pixel_mask_value = self.bad_pixel_mask_value
